#include <winsock2.h>
#include <windows.h>

#ifdef UNDER_CE

#include <shellapi.h>
#include <errno.h>
#include "wincompat.h"
#include "sys/stat.h"

int errno;

static BOOL AsciiToUnicode(const char* src, WCHAR* dest, int *pcchWideChar);
static wchar_t* StrDupAsciiToUnicode(const char* str);
static BOOL UnicodeToAscii(const wchar_t* src, char* dest, int *pcchChar);
static time_t ConvertWindowsToUnixTime(FILETIME* ft); // used by fstat


static BOOL AsciiToUnicode(const char* src, WCHAR* dest, int *pcchWideChar)
{
    int destSize;

    if (!pcchWideChar)
    {
        return FALSE;
    }

    destSize = MultiByteToWideChar(CP_ACP, 0, src, -1, dest, *pcchWideChar);
    if (destSize <= 0)
    {
        return FALSE;
    }

    *pcchWideChar = destSize;
    return TRUE;
}


static wchar_t* StrDupAsciiToUnicode(const char* str)
{
    wchar_t *result;
    int destSize = 0;
    if (FALSE == AsciiToUnicode(str, NULL, &destSize))
    {
        return NULL;
    }

    result = (wchar_t*)malloc(destSize * sizeof(wchar_t));
    if (!result)
    {
        return NULL;
    }

    return AsciiToUnicode(str, result, &destSize) ? result : NULL;
}


static BOOL UnicodeToAscii(const wchar_t* src, char* dest, int *pcchChar)
{
    int destSize;

    if (!pcchChar)
    {
        return FALSE;
    }

    destSize = WideCharToMultiByte(CP_ACP, 0, src, -1, dest, *pcchChar, NULL, NULL);
    if (destSize <= 0)
    {
        return FALSE;
    }
    
    *pcchChar = destSize;
    return TRUE;
}



static time_t ConvertWindowsToUnixTime(FILETIME* ft)
{
    LONGLONG ll = ft->dwHighDateTime;
    ll <<= 32;
    ll |= ft->dwLowDateTime;

    ll -= 116444736000000000;
    if (ll < 0)
    {
        errno = -1;
        return -1;
    }

    ll /= 10000000;

    return ll;
}


void ce_abort(char* file, int line)
{
    RETAILMSG(1, (L"!!! ABORT !!! file %S line %d\r\n", file, line));
    ASSERT(0);
    exit(-1);
}


int _close(int fd)
{
    return closesocket(fd);
}


int _read(int fd, void *buffer, unsigned int count)
{
    DWORD dwRead = 0;
    if (ReadFile((HANDLE)fd, buffer, count, &dwRead, NULL))
    {
        return dwRead;
    }
    else
    {
        return -1;
    }	
}


int _write(int fd, const void *buffer, unsigned int count)
{
    DWORD dwWritten = 0;
    if (WriteFile((HANDLE)fd, buffer, count, &dwWritten, NULL))
    {
        return dwWritten;
    }
    else
    {
        return -1;
    }	
}


int _unlink(const char* pathname)
{
    BOOL bSuccess;
    WCHAR* wPathName = StrDupAsciiToUnicode(pathname);
    bSuccess = DeleteFileW(wPathName);
    free(wPathName);	
    if (bSuccess)
    {
        return 0;	
    }
    else
    {			
        return -1;
    }	
}


int _mkdir(const char *pathname)
{

    BOOL bRet;
    wchar_t* wzDirectoryName = StrDupAsciiToUnicode(pathname);
    if (!wzDirectoryName)
    {
        errno = EINVAL;
        return -1;
    }

    bRet = CreateDirectoryW(wzDirectoryName, NULL);
    free(wzDirectoryName);

    if (bRet)
    {
        return 0;
    }
    switch(GetLastError())
    {
    case ERROR_PATH_NOT_FOUND:
        errno = ENOENT;
        break;
    default:
        errno = 0;
        break;
    }
    return -1;
}


int _access(const char *path, int mode)
{
    SHFILEINFOW fi;
    DWORD_PTR ret;
    wchar_t *wzPath;

    wzPath = StrDupAsciiToUnicode(path);
    if (!wzPath)
    {
        errno = EINVAL;
        return -1;
    }

    ret = SHGetFileInfo( wzPath, 0, &fi, sizeof(fi), SHGFI_ATTRIBUTES );
    free(wzPath);

    if (0 == ret)
    {
        errno = ENOENT;
        return -1;
    }
    // existence ?
    if( mode == 0 )
        return 0;
    // write permission ?
    if( mode & 2 )
    {
        if( fi.dwAttributes & SFGAO_READONLY )
        {
            errno = EACCES;
            return -1;
        }
    }
    return 0;
}


int fstat(FILE *desc, struct stat* st)
{
    HANDLE h;
    BY_HANDLE_FILE_INFORMATION info;

    h = (HANDLE) desc;
    if (h == INVALID_HANDLE_VALUE || st == NULL)
    {
        return -1;
    }

    if (!GetFileInformationByHandle(h, &info))
    {
        return -1;
    }

    // Set read/write/execute flags.
    st->st_mode = 
        S_IRUSR | S_IROTH | S_IRGRP |
        S_IWUSR | S_IWGRP | S_IWOTH |
        S_IXUSR | S_IXGRP | S_IXOTH;

    // FILE_ATTRIBUTE_DIRECTORY: Indicates that the handle identifies a directory.
    if (info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
        st->st_mode |= S_IFDIR;
    }
    else
    {
        st->st_mode |= S_IFREG;		
    }

    // FILE_ATTRIBUTE_READONLY:
    //		Indicates that the file or directory is read-only. 
    //
    // FILE_ATTRIBUTE_INROM:
    //		Indicates that the file is an operating system file stored in ROM.
    //		These files are read-only. They cannot be modified.
    //
    // FILE_ATTRIBUTE_ROMMODULE:
    //		Indicates that this file is an operating system file stored in ROM,
    //		The CreateFile function cannot be used to access this file.
    //		Instead the LoadLibrary and the CreateProcess functions must be used.        
    if (info.dwFileAttributes & (FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_INROM | FILE_ATTRIBUTE_ROMMODULE))
    {
        // Remove write flags
        st->st_mode &= ~(S_IWUSR | S_IWGRP | S_IWOTH);
    }

    if (info.dwFileAttributes & FILE_ATTRIBUTE_ROMMODULE)
    {
        // Remove read flags
        st->st_mode &= ~(S_IRUSR | S_IROTH | S_IRGRP);
    }

    st->st_size = info.nFileSizeLow;
    st->st_atime = ConvertWindowsToUnixTime(&info.ftLastAccessTime);
    st->st_mtime = ConvertWindowsToUnixTime(&info.ftLastWriteTime);
    st->st_ctime = ConvertWindowsToUnixTime(&info.ftCreationTime);
    st->st_dev = 0;
    st->st_ino = 0;
    st->st_uid = 0;
    st->st_gid = 0;
    st->st_rdev = 0;	
    st->st_nlink = 1;

    return 0;
}


struct tm *localtime(const time_t *time)
{
    // Only used by standard log.
    // We can provide our own log call back.
    return NULL;
}


//struct tm *localtime_r(const time_t *time, struct tm *result)
//{
//    return _localtime64_s(result, time) ? NULL : result;
//}


size_t strftime(char *s, size_t maxs, const char *f, const struct tm *t)
{
    size_t ret = 0;
    wchar_t *destW = NULL;
    wchar_t *formatW = NULL;
    int cchChar = maxs;

    destW = (wchar_t*)malloc(maxs * sizeof(wchar_t));
    if (!destW)
    {
        goto cleanup;
    }

    formatW = StrDupAsciiToUnicode(f);
    if (!formatW)
    {
        goto cleanup;
    }

    ret = wcsftime(destW, maxs, formatW, t);
    if (0 == ret)
    {
        goto cleanup;
    }

    if (!UnicodeToAscii(destW, s, &cchChar))
    {
        ret = 0;
    }

cleanup:
    free(destW);
    free(formatW);

    return ret;
}


int WSAStringToAddressA(LPSTR               AddressString,
                        INT                 AddressFamily,
                        LPWSAPROTOCOL_INFOA lpProtocolInfo,
                        LPSOCKADDR          lpAddress,
                        LPINT               lpAddressLength
                        )
{
    int ret;
    wchar_t* wzAddress;

    if (lpProtocolInfo)
    {
        errno = EINVAL;
        return -1;
    }

    wzAddress = StrDupAsciiToUnicode(AddressString);
    if (!wzAddress)
    {
        errno = EINVAL;
        return -1;
    }

    ret = WSAStringToAddressW(wzAddress, AddressFamily, NULL, lpAddress, lpAddressLength);
    free(wzAddress);
    return ret;
}


BOOL SHGetSpecialFolderPathA(HWND hwnd, __out_ecount(MAX_PATH) LPSTR pszPath, int csidl, BOOL fCreate)
{
    wchar_t buf[ MAX_PATH ];
    int len = MAX_PATH;

    return SHGetSpecialFolderPath( hwnd, buf, csidl, fCreate ) && UnicodeToAscii( buf, pszPath, &len );
}

BOOL GetUserNameA(char* lpBuffer, LPDWORD lpnSize)
{
    // Returns the local name passed to the SetCurrentUser function.
    static const EXTENDED_NAME_FORMAT nameFormat = NameWindowsCeLocal;

    BOOL ret;
    wchar_t* tmp;
    DWORD length;
    
    if (!lpBuffer)
    {
        return GetUserNameExW(nameFormat, NULL, lpnSize);
    }

    length = *lpnSize;
    tmp = (wchar_t*)malloc(length * sizeof(wchar_t));
    if (!tmp)
    {
        return FALSE;
    }

    ret = GetUserNameExW(nameFormat, tmp, &length);
    if (!ret)
    {
        goto cleanup;
    }

    ret = UnicodeToAscii(tmp, lpBuffer, lpnSize);
    if (ret && (*lpnSize > 0))
    {
        // Do not count null-terminator.
        --(*lpnSize);
    }

cleanup:
    free(tmp);
    return ret;
}

HANDLE GetStdHandle(DWORD nStdHandle)
{
    // TODO : Review
    return NULL;
}

BOOL GetConsoleMode(HANDLE hConsoleHandle, LPDWORD lpMode)
{
    // TODO : Review
    if (!lpMode)
    {
        return FALSE;
    }
    *lpMode = 0;
    return TRUE;
}

BOOL SetConsoleMode(HANDLE hConsoleHandle, DWORD dwMode)
{
    // TODO : Review
    return TRUE;
}

#endif


#define QUOTE(str) #str
#define EXPAND_AND_QUOTE(str) QUOTE(str)

const char* ssh_strerror(int err)
{
	typedef struct
	{
		int err;
		const char* descr;
	} err_info;

	static const err_info einfos[] = {
		{ WSAEINTR, "WSAEINTR (" EXPAND_AND_QUOTE(WSAEINTR) ")" },
		{ WSAEBADF, "WSAEBADF (10009)" },
		{ WSAEACCES, "WSAEACCES (10013)" },
		{ WSAEFAULT, "WSAEFAULT                        10014" },
		{ WSAEINVAL, "WSAEINVAL                        10022" },
		{ WSAEMFILE, "WSAEMFILE                        10024" },
		{ WSAEWOULDBLOCK, "WSAEWOULDBLOCK                   10035" },
		{ WSAEINPROGRESS, "WSAEINPROGRESS                   10036" },
		{ WSAEALREADY, "WSAEALREADY                      10037" },
		{ WSAENOTSOCK, "WSAENOTSOCK                      10038" },
		{ WSAEDESTADDRREQ, "WSAEDESTADDRREQ                  10039" },
		{ WSAEMSGSIZE, "WSAEMSGSIZE                      10040" },
		{ WSAEPROTOTYPE, "WSAEPROTOTYPE                    10041" },
		{ WSAENOPROTOOPT, "WSAENOPROTOOPT                   10042" },
		{ WSAEPROTONOSUPPORT, "WSAEPROTONOSUPPORT               10043" },
		{ WSAESOCKTNOSUPPORT, "WSAESOCKTNOSUPPORT               10044" },
		{ WSAEOPNOTSUPP, "WSAEOPNOTSUPP                    10045" },
		{ WSAEPFNOSUPPORT, "WSAEPFNOSUPPORT                  10046" },
		{ WSAEAFNOSUPPORT, "WSAEAFNOSUPPORT                  10047" },
		{ WSAEADDRINUSE, "WSAEADDRINUSE                    10048" },
		{ WSAEADDRNOTAVAIL, "WSAEADDRNOTAVAIL                 10049" },
		{ WSAENETDOWN, "WSAENETDOWN                      10050" },
		{ WSAENETUNREACH, "WSAENETUNREACH                   10051" },
		{ WSAENETRESET, "WSAENETRESET                     10052" },
		{ WSAECONNABORTED, "WSAECONNABORTED                  10053" },
		{ WSAECONNRESET, "WSAECONNRESET                    10054" },
		{ WSAENOBUFS, "WSAENOBUFS                       10055" },
		{ WSAEISCONN, "WSAEISCONN                       10056" },
		{ WSAENOTCONN, "WSAENOTCONN                      10057" },
		{ WSAESHUTDOWN, "WSAESHUTDOWN                     10058" },
		{ WSAETOOMANYREFS, "WSAETOOMANYREFS                  10059" },
		{ WSAETIMEDOUT, "WSAETIMEDOUT                     10060" },
		{ WSAECONNREFUSED, "WSAECONNREFUSED                  10061" },
		{ WSAELOOP, "WSAELOOP                         10062" },
		{ WSAENAMETOOLONG, "WSAENAMETOOLONG                  10063" },
		{ WSAEHOSTDOWN, "WSAEHOSTDOWN                     10064" },
		{ WSAEHOSTUNREACH, "WSAEHOSTUNREACH                  10065" },
		{ WSAENOTEMPTY, "WSAENOTEMPTY                     10066" },
		{ WSAEPROCLIM, "WSAEPROCLIM                      10067" },
		{ WSAEUSERS, "WSAEUSERS                        10068" },
		{ WSAEDQUOT, "WSAEDQUOT                        10069" },
		{ WSAESTALE, "WSAESTALE                        10070" },
		{ WSAEREMOTE, "WSAEREMOTE                       10071" },
		{ WSASYSNOTREADY, "WSASYSNOTREADY                   10091" },
		{ WSAVERNOTSUPPORTED, "WSAVERNOTSUPPORTED               10092" },
		{ WSANOTINITIALISED, "WSANOTINITIALISED                10093" },
		{ WSAEDISCON, "WSAEDISCON                       10101" },
		{ WSAENOMORE, "WSAENOMORE                       10102" },
		{ WSAECANCELLED, "WSAECANCELLED                    10103" },
		{ WSAEINVALIDPROCTABLE, "WSAEINVALIDPROCTABLE             10104" },
		{ WSAEINVALIDPROVIDER, "WSAEINVALIDPROVIDER              10105" },
		{ WSAEPROVIDERFAILEDINIT, "WSAEPROVIDERFAILEDINIT           10106" },
		{ WSASYSCALLFAILURE, "WSASYSCALLFAILURE                10107" },
		{ WSASERVICE_NOT_FOUND, "WSASERVICE_NOT_FOUND             10108" },
		{ WSATYPE_NOT_FOUND, "WSATYPE_NOT_FOUND                10109" },
		{ WSA_E_NO_MORE, "WSA_E_NO_MORE                    10110" },
		{ WSA_E_CANCELLED, "WSA_E_CANCELLED                  10111" },
		{ WSAEREFUSED, "WSAEREFUSED                      10112" },
		{ WSAHOST_NOT_FOUND, "WSAHOST_NOT_FOUND                11001" },
		{ WSATRY_AGAIN, "WSATRY_AGAIN                     11002" },
		{ WSANO_RECOVERY, "WSANO_RECOVERY                   11003" },
		{ WSANO_DATA, "WSANO_DATA                       11004" },
		{ WSA_QOS_RECEIVERS, "WSA_QOS_RECEIVERS                11005" },
		{ WSA_QOS_SENDERS, "WSA_QOS_SENDERS                  11006" },
		{ WSA_QOS_NO_SENDERS, "WSA_QOS_NO_SENDERS               11007" },
		{ WSA_QOS_NO_RECEIVERS, "WSA_QOS_NO_RECEIVERS             11008" },
		{ WSA_QOS_REQUEST_CONFIRMED, "WSA_QOS_REQUEST_CONFIRMED        11009" },
		{ WSA_QOS_ADMISSION_FAILURE, "WSA_QOS_ADMISSION_FAILURE        11010" },
		{ WSA_QOS_POLICY_FAILURE, "WSA_QOS_POLICY_FAILURE           11011" },
		{ WSA_QOS_BAD_STYLE, "WSA_QOS_BAD_STYLE                11012" },
		{ WSA_QOS_BAD_OBJECT, "WSA_QOS_BAD_OBJECT               11013" },
		{ WSA_QOS_TRAFFIC_CTRL_ERROR, "WSA_QOS_TRAFFIC_CTRL_ERROR       11014" },
		{ WSA_QOS_GENERIC_ERROR, "WSA_QOS_GENERIC_ERROR            11015" },
		{ WSA_QOS_ESERVICETYPE, "WSA_QOS_ESERVICETYPE             11016" },
		{ WSA_QOS_EFLOWSPEC, "WSA_QOS_EFLOWSPEC                11017" },
		{ WSA_QOS_EPROVSPECBUF, "WSA_QOS_EPROVSPECBUF             11018" },
		{ WSA_QOS_EFILTERSTYLE, "WSA_QOS_EFILTERSTYLE             11019" },
		{ WSA_QOS_EFILTERTYPE, "WSA_QOS_EFILTERTYPE              11020" },
		{ WSA_QOS_EFILTERCOUNT, "WSA_QOS_EFILTERCOUNT             11021" },
		{ WSA_QOS_EOBJLENGTH, "WSA_QOS_EOBJLENGTH               11022" },
		{ WSA_QOS_EFLOWCOUNT, "WSA_QOS_EFLOWCOUNT               11023" },
		{ WSA_QOS_EUNKOWNPSOBJ, "WSA_QOS_EUNKOWNPSOBJ             11024" },
		{ WSA_QOS_EPOLICYOBJ, "WSA_QOS_EPOLICYOBJ               11025" },
		{ WSA_QOS_EFLOWDESC, "WSA_QOS_EFLOWDESC                11026" },
		{ WSA_QOS_EPSFLOWSPEC, "WSA_QOS_EPSFLOWSPEC              11027" },
		{ WSA_QOS_EPSFILTERSPEC, "WSA_QOS_EPSFILTERSPEC            11028" },
		{ WSA_QOS_ESDMODEOBJ, "WSA_QOS_ESDMODEOBJ               11029" },
		{ WSA_QOS_ESHAPERATEOBJ, "WSA_QOS_ESHAPERATEOBJ            11030" }
	};

	return einfos[0].descr;
}
